-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------


-- Purpose:
--  1) Verify unb1_board_mesh_reorder_bidir for SOSI using the unb1_board_mesh_model_sosi
--  2) Verify unb1_board_mesh_reorder_bidir for SISO using the unb1_board_mesh_model_siso
--  3) Verify the serial unb1_board_mesh_model_sl
-- Usage:
--  > do wave_unb1_board_mesh_reorder_bidir.do
--  > run -all
--  1) Self check SOSI should report no bn_rx_valid[] or fn_rx_valid[] error
--  2) Self check SISO should report no bn_tx_ready[] or fn_tx_ready[] error
--  3) Self check unb1_board_mesh_model_sl operates at PHY level using the
--     SOSI.valid as stimuli and the SOSI.valid result as reference and should
--     report no error.
--  . Manually observe that bn_rx_valid and fn_rx_valid show a series of pulses
-- Description:
--  . See unb1_board_mesh_reorder_bidir.vhd
--  . Use c_reorder=FALSE for a default FN to BN and BN to BN transpose without
--    reordering on the PCB. Default use c_reorder=TRUE to also model the
--    transceiver bus reorderings on UniBoard PCB.


LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.unb1_board_pkg.ALL;
USE work.tb_unb1_board_pkg.ALL;


ENTITY tb_unb1_board_mesh_reorder_bidir IS
END tb_unb1_board_mesh_reorder_bidir;


ARCHITECTURE tb OF tb_unb1_board_mesh_reorder_bidir IS

  CONSTANT c_reorder      : BOOLEAN := TRUE;
  
  CONSTANT c_chip_id_w    : NATURAL := c_unb1_board_aux.chip_id_w;     -- = 3 to fit 8 fpgas in [2:0]
  CONSTANT c_nof_node     : NATURAL := c_unb1_board_nof_node;          -- = 4
  CONSTANT c_nof_bus      : NATURAL := c_unb1_board_tr.nof_bus;        -- = 4
  CONSTANT c_bus_w        : NATURAL := c_unb1_board_tr.bus_w;          -- = 4
  
  CONSTANT c_repeat       : NATURAL := 3;
  CONSTANT c_clk_period   : TIME := 10 ns;
  
  SIGNAL tb_end                 : STD_LOGIC := '0';
  SIGNAL clk                    : STD_LOGIC := '0';
  
  ------------------------------------------------------------------------------
  -- SL
  ------------------------------------------------------------------------------
  -- . PHY signals to mesh
  SIGNAL fn_tx_phy_sl_3arr      : t_unb1_board_mesh_sl_3arr;
  SIGNAL bn_tx_phy_sl_3arr      : t_unb1_board_mesh_sl_3arr;
  
  -- . PHY signals from mesh
  SIGNAL fn_rx_phy_sl_3arr      : t_unb1_board_mesh_sl_3arr;
  SIGNAL fn_rx_phy_sl           : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL fn_rx_phy_valid        : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_rx_phy_sl_3arr      : t_unb1_board_mesh_sl_3arr;
  SIGNAL bn_rx_phy_sl           : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_rx_phy_valid        : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  
  ------------------------------------------------------------------------------
  -- SOSI
  ------------------------------------------------------------------------------
  -- . Monitoring signals
  SIGNAL fn_tx_valid            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_tx_valid            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL fn_rx_valid            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_rx_valid            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);

  -- . User domain driving stimuli signals
  SIGNAL fn_tx_usr_sosi_3arr    : t_unb1_board_mesh_sosi_3arr := (OTHERS=>(OTHERS=>(OTHERS=>c_dp_sosi_rst)));
  SIGNAL bn_tx_usr_sosi_3arr    : t_unb1_board_mesh_sosi_3arr := (OTHERS=>(OTHERS=>(OTHERS=>c_dp_sosi_rst)));
  
  -- . PHY signals to mesh
  SIGNAL fn_tx_phy_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  SIGNAL bn_tx_phy_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  
  -- . PHY signals from mesh
  SIGNAL fn_rx_phy_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  SIGNAL bn_rx_phy_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  
  -- . User domain result signals
  SIGNAL fn_rx_usr_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  SIGNAL bn_rx_usr_sosi_3arr    : t_unb1_board_mesh_sosi_3arr;
  
  ------------------------------------------------------------------------------
  -- SISO
  ------------------------------------------------------------------------------
  -- . Monitoring signals
  SIGNAL fn_rx_ready            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_rx_ready            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL fn_tx_ready            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);
  SIGNAL bn_tx_ready            : STD_LOGIC_VECTOR(c_nof_node*c_nof_bus*c_bus_w-1 DOWNTO 0);

  -- . User domain result signals
  SIGNAL fn_tx_usr_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  SIGNAL bn_tx_usr_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  
  -- . PHY signals from mesh
  SIGNAL fn_tx_phy_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  SIGNAL bn_tx_phy_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  
  -- . PHY signals to mesh
  SIGNAL fn_rx_phy_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  SIGNAL bn_rx_phy_siso_3arr    : t_unb1_board_mesh_siso_3arr;
  
  -- . User domain driving stimuli signals
  SIGNAL fn_rx_usr_siso_3arr    : t_unb1_board_mesh_siso_3arr := (OTHERS=>(OTHERS=>(OTHERS=>c_dp_siso_rst)));
  SIGNAL bn_rx_usr_siso_3arr    : t_unb1_board_mesh_siso_3arr := (OTHERS=>(OTHERS=>(OTHERS=>c_dp_siso_rst)));
  
BEGIN

  clk <= (NOT clk) OR tb_end AFTER c_clk_period/2;
  
  ------------------------------------------------------------------------------
  -- Stimuli
  ------------------------------------------------------------------------------
  
  p_stimuli : PROCESS
  BEGIN
    proc_common_wait_some_cycles(clk, 100);
    FOR R IN 0 TO c_repeat-1 LOOP
      -- Sequentially set a '1' pulse on every sosi.valid and and every
      -- siso.ready then check that these get accross the UniBoard mesh in the
      -- correct order by comparing the output pattern with the (delayed)
      -- input pattern.
      
      -- FN
      FOR I IN 0 TO c_nof_node-1 LOOP
        FOR J IN 0 TO c_nof_bus-1 LOOP
          FOR K IN 0 TO c_bus_w-1 LOOP
            fn_tx_usr_sosi_3arr(I)(J)(K).valid <= '1';
            fn_rx_usr_siso_3arr(I)(J)(K).ready <= '1';
            proc_common_wait_some_cycles(clk, 1);
            fn_tx_usr_sosi_3arr(I)(J)(K).valid <= '0';
            fn_rx_usr_siso_3arr(I)(J)(K).ready <= '0';
          END LOOP;
        END LOOP;
      END LOOP;
      
      -- BN
      FOR I IN 0 TO c_nof_node-1 LOOP
        FOR J IN 0 TO c_nof_bus-1 LOOP
          FOR K IN 0 TO c_bus_w-1 LOOP
            bn_tx_usr_sosi_3arr(I)(J)(K).valid <= '1';
            bn_rx_usr_siso_3arr(I)(J)(K).ready <= '1';
            proc_common_wait_some_cycles(clk, 1);
            bn_tx_usr_sosi_3arr(I)(J)(K).valid <= '0';
            bn_rx_usr_siso_3arr(I)(J)(K).ready <= '0';
          END LOOP;
        END LOOP;
      END LOOP;
      
    END LOOP;
    
    proc_common_wait_some_cycles(clk, 100);
    tb_end <= '1';    
    WAIT;
  END PROCESS;
  
  
  ------------------------------------------------------------------------------
  -- Monitor and verify
  ------------------------------------------------------------------------------
  
  gen_node : FOR I IN 0 TO c_nof_node-1 GENERATE
    gen_bus : FOR J IN 0 TO c_nof_bus-1 GENERATE
      gen_lanes : FOR K IN c_bus_w-1 DOWNTO 0 GENERATE
        -- SOSI
        -- . Transmit order
        fn_tx_valid((I*c_nof_bus+J)*c_bus_w+K) <= fn_tx_usr_sosi_3arr(I)(J)(K).valid;  -- FN
        bn_tx_valid((I*c_nof_bus+J)*c_bus_w+K) <= bn_tx_usr_sosi_3arr(I)(J)(K).valid;  -- BN
        -- . Receive order (compensate for the transpose)
        fn_rx_valid((J*c_nof_bus+I)*c_bus_w+K) <= fn_rx_usr_sosi_3arr(I)(J)(K).valid;  -- FN
        bn_rx_valid((J*c_nof_bus+I)*c_bus_w+K) <= bn_rx_usr_sosi_3arr(I)(J)(K).valid;  -- BN
        
        -- SISO
        -- . Transmit order
        fn_rx_ready((I*c_nof_bus+J)*c_bus_w+K) <= fn_rx_usr_siso_3arr(I)(J)(K).ready;  -- FN
        bn_rx_ready((I*c_nof_bus+J)*c_bus_w+K) <= bn_rx_usr_siso_3arr(I)(J)(K).ready;  -- BN
        -- . Receive order (compensate for the transpose)
        fn_tx_ready((J*c_nof_bus+I)*c_bus_w+K) <= fn_tx_usr_siso_3arr(I)(J)(K).ready;  -- FN
        bn_tx_ready((J*c_nof_bus+I)*c_bus_w+K) <= bn_tx_usr_siso_3arr(I)(J)(K).ready;  -- BN
      END GENERATE;
    END GENERATE;
  END GENERATE;
  
  p_verify : PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      -- SOSI
      ASSERT fn_tx_valid = bn_rx_valid REPORT "Mesh bn_rx_valid error" SEVERITY ERROR;
      ASSERT bn_tx_valid = fn_rx_valid REPORT "Mesh fn_rx_valid error" SEVERITY ERROR;
      -- SISO
      ASSERT fn_rx_ready = bn_tx_ready REPORT "Mesh bn_tx_ready error" SEVERITY ERROR;
      ASSERT bn_rx_ready = fn_tx_ready REPORT "Mesh fn_tx_ready error" SEVERITY ERROR;
    END IF;
  END PROCESS;
  
  
  ------------------------------------------------------------------------------
  -- UniBoard FN0,1,2,3
  ------------------------------------------------------------------------------
  
  gen_fn : FOR I IN 0 TO c_nof_node-1 GENERATE
    u_order : ENTITY work.unb1_board_mesh_reorder_bidir
    GENERIC MAP (
      g_node_type => e_fn,
      g_reorder   => c_reorder
    )
    PORT MAP (
      chip_id          => TO_UVEC(I, c_chip_id_w),  -- chip id 0, 1, 2, 3
      
      -- Transmit clock domain
      tx_clk           => clk,
      tx_usr_sosi_2arr => fn_tx_usr_sosi_3arr(I),  -- user sosi to phy = sosi.valid driver from FN user
      tx_usr_siso_2arr => fn_tx_usr_siso_3arr(I),  -- user siso from phy = siso.ready result to FN user
      tx_phy_sosi_2arr => fn_tx_phy_sosi_3arr(I),  -- phy sosi to mesh
      tx_phy_siso_2arr => fn_tx_phy_siso_3arr(I),  -- phy siso from mesh
      
      -- Receive clock domain
      rx_clk           => clk,
      rx_phy_sosi_2arr => fn_rx_phy_sosi_3arr(I),  -- phy sosi from mesh
      rx_phy_siso_2arr => fn_rx_phy_siso_3arr(I),  -- phy siso to mesh
      rx_usr_sosi_2arr => fn_rx_usr_sosi_3arr(I),  -- user sosi from phy = sosi.valid result to FN user
      rx_usr_siso_2arr => fn_rx_usr_siso_3arr(I)   -- user siso to phy = siso.ready driver from FN user
    );
  END GENERATE;
  
  
  ------------------------------------------------------------------------------
  -- Mesh between FN0,1,2,3 and BN0,1,2,3 as it is wired on the UniBoard PCB
  ------------------------------------------------------------------------------

  -- >>> unb1_board_mesh_model_sl
  
  -- Use tx_phy SOSI.valid stimuli for input to unb1_board_mesh_model_sl
  gen_tx_serial : FOR I IN 0 TO c_nof_node-1 GENERATE
    gen_bus : FOR J IN 0 TO c_nof_bus-1 GENERATE
      gen_lanes : FOR K IN c_bus_w-1 DOWNTO 0 GENERATE
        fn_tx_phy_sl_3arr(I)(J)(K) <= fn_tx_phy_sosi_3arr(I)(J)(K).valid;
        bn_tx_phy_sl_3arr(I)(J)(K) <= bn_tx_phy_sosi_3arr(I)(J)(K).valid;
      END GENERATE;
    END GENERATE;
  END GENERATE;
    
  u_pcb_mesh_serial : ENTITY work.unb1_board_mesh_model_sl
  GENERIC MAP (
    g_reorder => c_reorder
  )
  PORT MAP (
    -- FN to BN
    fn_tx_sl_3arr  => fn_tx_phy_sl_3arr,
    bn_rx_sl_3arr  => bn_rx_phy_sl_3arr,
    
    -- BN to FN
    bn_tx_sl_3arr  => bn_tx_phy_sl_3arr,
    fn_rx_sl_3arr  => fn_rx_phy_sl_3arr
  );
  
  -- Use rx_phy SOSI.valid as reference output to verify output of unb1_board_mesh_model_sl
  mon_rx_serial : FOR I IN 0 TO c_nof_node-1 GENERATE
    gen_bus : FOR J IN 0 TO c_nof_bus-1 GENERATE
      gen_lanes : FOR K IN c_bus_w-1 DOWNTO 0 GENERATE
        -- Monitor SOSI valids in SLV
        bn_rx_phy_valid((I*c_nof_bus+J)*c_bus_w+K) <= bn_rx_phy_sosi_3arr(I)(J)(K).valid;
        fn_rx_phy_valid((I*c_nof_bus+J)*c_bus_w+K) <= fn_rx_phy_sosi_3arr(I)(J)(K).valid;
        -- Monitor SL valids in SLV
        bn_rx_phy_sl((I*c_nof_bus+J)*c_bus_w+K) <= bn_rx_phy_sl_3arr(I)(J)(K);
        fn_rx_phy_sl((I*c_nof_bus+J)*c_bus_w+K) <= fn_rx_phy_sl_3arr(I)(J)(K);
      END GENERATE;
    END GENERATE;
  END GENERATE;

  p_verify_serial : PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      ASSERT bn_rx_phy_sl=bn_rx_phy_valid REPORT "unb1_board_mesh_model_sl FN->BN error" SEVERITY ERROR;
      ASSERT fn_rx_phy_sl=fn_rx_phy_valid REPORT "unb1_board_mesh_model_sl BN->FN error" SEVERITY ERROR;
    END IF;
  END PROCESS;
  
  
  -- >>> unb1_board_mesh_model_sosi
  
  u_pcb_mesh_sosi : ENTITY work.unb1_board_mesh_model_sosi
  GENERIC MAP (
    g_reorder => c_reorder
  )
  PORT MAP (
    -- FN to BN
    fn0_tx_sosi_2arr  => fn_tx_phy_sosi_3arr(0),
    fn1_tx_sosi_2arr  => fn_tx_phy_sosi_3arr(1),
    fn2_tx_sosi_2arr  => fn_tx_phy_sosi_3arr(2),
    fn3_tx_sosi_2arr  => fn_tx_phy_sosi_3arr(3),
    
    bn0_rx_sosi_2arr  => bn_rx_phy_sosi_3arr(0),
    bn1_rx_sosi_2arr  => bn_rx_phy_sosi_3arr(1),
    bn2_rx_sosi_2arr  => bn_rx_phy_sosi_3arr(2),
    bn3_rx_sosi_2arr  => bn_rx_phy_sosi_3arr(3),
    
    -- BN to FN
    bn0_tx_sosi_2arr  => bn_tx_phy_sosi_3arr(0),
    bn1_tx_sosi_2arr  => bn_tx_phy_sosi_3arr(1),
    bn2_tx_sosi_2arr  => bn_tx_phy_sosi_3arr(2),
    bn3_tx_sosi_2arr  => bn_tx_phy_sosi_3arr(3),
    
    fn0_rx_sosi_2arr  => fn_rx_phy_sosi_3arr(0),
    fn1_rx_sosi_2arr  => fn_rx_phy_sosi_3arr(1),
    fn2_rx_sosi_2arr  => fn_rx_phy_sosi_3arr(2),
    fn3_rx_sosi_2arr  => fn_rx_phy_sosi_3arr(3)
  );
  
  
  -- >>> unb1_board_mesh_model_siso
  
  u_pcb_mesh_siso : ENTITY work.unb1_board_mesh_model_siso
  GENERIC MAP (
    g_reorder => c_reorder
  )
  PORT MAP (
    -- FN to BN
    fn0_rx_siso_2arr  => fn_rx_phy_siso_3arr(0),
    fn1_rx_siso_2arr  => fn_rx_phy_siso_3arr(1),
    fn2_rx_siso_2arr  => fn_rx_phy_siso_3arr(2),
    fn3_rx_siso_2arr  => fn_rx_phy_siso_3arr(3),
    
    bn0_tx_siso_2arr  => bn_tx_phy_siso_3arr(0),
    bn1_tx_siso_2arr  => bn_tx_phy_siso_3arr(1),
    bn2_tx_siso_2arr  => bn_tx_phy_siso_3arr(2),
    bn3_tx_siso_2arr  => bn_tx_phy_siso_3arr(3),
    
    -- BN to FN
    bn0_rx_siso_2arr  => bn_rx_phy_siso_3arr(0),
    bn1_rx_siso_2arr  => bn_rx_phy_siso_3arr(1),
    bn2_rx_siso_2arr  => bn_rx_phy_siso_3arr(2),
    bn3_rx_siso_2arr  => bn_rx_phy_siso_3arr(3),
    
    fn0_tx_siso_2arr  => fn_tx_phy_siso_3arr(0),
    fn1_tx_siso_2arr  => fn_tx_phy_siso_3arr(1),
    fn2_tx_siso_2arr  => fn_tx_phy_siso_3arr(2),
    fn3_tx_siso_2arr  => fn_tx_phy_siso_3arr(3)
  );
  
  
  ------------------------------------------------------------------------------
  -- UniBoard BN0,1,2,3
  ------------------------------------------------------------------------------
  
  gen_bn : FOR I IN 0 TO c_nof_node-1 GENERATE
    u_order : ENTITY work.unb1_board_mesh_reorder_bidir
    GENERIC MAP (
      g_node_type => e_bn,
      g_reorder   => c_reorder
    )
    PORT MAP (
      chip_id          => TO_UVEC(c_nof_node+I, c_chip_id_w),  -- chip id 4, 5, 6, 7
      
      -- Transmit clock domain
      tx_clk           => clk,
      tx_usr_sosi_2arr => bn_tx_usr_sosi_3arr(I),  -- user sosi to phy = sosi.valid driver from BN user
      tx_usr_siso_2arr => bn_tx_usr_siso_3arr(I),  -- user siso from phy = siso.ready result to BN user
      tx_phy_sosi_2arr => bn_tx_phy_sosi_3arr(I),  -- phy sosi to mesh
      tx_phy_siso_2arr => bn_tx_phy_siso_3arr(I),  -- phy siso from mesh
                                                                         
      -- Receive clock domain                                            
      rx_clk           => clk,                                           
      rx_phy_sosi_2arr => bn_rx_phy_sosi_3arr(I),  -- phy sosi from mesh
      rx_phy_siso_2arr => bn_rx_phy_siso_3arr(I),  -- phy siso to mesh
      rx_usr_sosi_2arr => bn_rx_usr_sosi_3arr(I),  -- user sosi from phy = sosi.valid result to BN user
      rx_usr_siso_2arr => bn_rx_usr_siso_3arr(I)   -- user siso to phy = siso.ready driver from BN user
    );
  END GENERATE;
  
END tb;
