-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

--  RO  read only  (no VHDL present to access HW in write mode)
--  WO  write only (no VHDL present to access HW in read mode)
--  WE  write event (=WO)
--  WR  write control, read control
--  RW  read status, write control  
--  RC  read, clear on read
--  FR  FIFO read
--  FW  FIFO write
--
--  wi  Bits    R/W Name          Default  Description                 |REG_EPCS|                      
--  =============================================================================
--  0   [31..0] WO  reconfigure   0x0
--  1   [2..0]  WO  param
--  2   [0]     WE  read_param
--  3   [0]     WE  write_param 
--  4   [23..0] RO  data_out
--  5   [23..0] WO  data_in
--  6   [0]     RO  busy
--  =============================================================================

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY remu_reg IS
  GENERIC (
    g_data_w : NATURAL := 24
  );
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;    
    mm_clk            : IN  STD_LOGIC;

    epcs_rst          : IN  STD_LOGIC;    
    epcs_clk          : IN  STD_LOGIC;
    
    -- Memory Mapped Slave 
    sla_in            : IN  t_mem_mosi; 
    sla_out           : OUT t_mem_miso; 
    
    reconfigure       : OUT STD_LOGIC;
    read_param        : OUT STD_LOGIC;
    write_param       : OUT STD_LOGIC;
    param             : OUT STD_LOGIC_VECTOR(2 DOWNTO 0);
    busy              : IN  STD_LOGIC;
    data_out          : IN  STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
    data_in           : OUT STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0)
    );
END remu_reg;


ARCHITECTURE rtl OF remu_reg IS

  CONSTANT c_mm_reg        : t_c_mem := (latency  => 1,
                                         adr_w    => ceil_log2(7),
                                         dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                         nof_dat  => 7,
                                         init_sl  => '0');                                               

  -- For safety, some commands require specific words to be written
  CONSTANT c_cmd_reconfigure : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0 ) := x"B007FAC7"; -- "Boot factory"

  SIGNAL mm_reconfigure      : STD_LOGIC;
  SIGNAL mm_read_param       : STD_LOGIC;
  SIGNAL mm_write_param      : STD_LOGIC;
  SIGNAL mm_param            : STD_LOGIC_VECTOR(2 DOWNTO 0);
  SIGNAL mm_busy             : STD_LOGIC;
  SIGNAL mm_data_out         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  SIGNAL mm_data_in          : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
     
BEGIN

  p_mm_reg : PROCESS (mm_rst, mm_clk)
  BEGIN
    IF mm_rst = '1' THEN
      -- Read access
      sla_out   <= c_mem_miso_rst;
      -- Write access, register values
      mm_reconfigure  <= '0';
      mm_read_param   <= '0';
      mm_write_param  <= '0';
      mm_param        <= (OTHERS=>'0');
      mm_data_in      <= (OTHERS=>'0');

    ELSIF rising_edge(mm_clk) THEN
      -- Read access defaults
      sla_out.rdval <= '0';
      
      -- Write event defaults
      mm_read_param  <= '0';
      mm_write_param <= '0';

      -- Write access: set register value
      IF sla_in.wr = '1' THEN
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          WHEN 0 => 
            IF sla_in.wrdata(c_word_w-1 DOWNTO 0) = c_cmd_reconfigure THEN
              mm_reconfigure <= '1';
            ELSE
              mm_reconfigure <= '0';
            END IF;
          WHEN 1 => 
            mm_param <= sla_in.wrdata(2 DOWNTO 0);
          WHEN 2 => 
            mm_read_param <= sla_in.wrdata(0);
          WHEN 3 => 
            mm_write_param <= sla_in.wrdata(0);
          WHEN 5 => 
            mm_data_in <= sla_in.wrdata(g_data_w-1 DOWNTO 0);
          WHEN OTHERS => NULL;  -- unused MM addresses
        END CASE;
        
      -- Read access: get register value
      ELSIF sla_in.rd = '1' THEN
        sla_out       <= c_mem_miso_rst;    -- set unused rddata bits to '0' when read
        sla_out.rdval <= '1';               -- c_mm_reg.latency = 1
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          WHEN 4 =>            
            sla_out.rddata(g_data_w-1 DOWNTO 0) <= mm_data_out;
          WHEN 6 =>  
            sla_out.rddata(0) <= mm_busy;
          WHEN OTHERS => NULL;  -- unused MM addresses
        END CASE;
      END IF;
    END IF;
  END PROCESS;

  u_spulse_write_param : ENTITY common_lib.common_spulse
  PORT MAP (
    in_rst    => mm_rst,
    in_clk    => mm_clk,
    in_pulse  => mm_write_param,
    in_busy   => OPEN,
    out_rst   => epcs_rst,
    out_clk   => epcs_clk,
    out_pulse => write_param
  );

  u_spulse_read_param : ENTITY common_lib.common_spulse
  PORT MAP (
    in_rst    => mm_rst,
    in_clk    => mm_clk,
    in_pulse  => mm_read_param,
    in_busy   => OPEN,
    out_rst   => epcs_rst,
    out_clk   => epcs_clk,
    out_pulse => read_param
  );

  u_spulse_reconfigure : ENTITY common_lib.common_spulse
  PORT MAP (
    in_rst    => mm_rst,
    in_clk    => mm_clk,
    in_pulse  => mm_reconfigure,
    in_busy   => OPEN,
    out_rst   => epcs_rst,
    out_clk   => epcs_clk,
    out_pulse => reconfigure
  );

  u_async_busy : ENTITY common_lib.common_async
  GENERIC MAP (
    g_rst_level => '0'
  )
  PORT MAP (
    rst  => mm_rst,
    clk  => mm_clk,
    din  => busy,
    dout => mm_busy
  );
    
  u_cross_param : ENTITY common_lib.common_reg_cross_domain
  PORT MAP (
    in_rst      => mm_rst,
    in_clk      => mm_clk,
    in_dat      => mm_param,
    in_done     => OPEN,
    out_rst     => epcs_rst,
    out_clk     => epcs_clk,
    out_dat     => param,
    out_new     => OPEN
  );

  u_cross_data_in : ENTITY common_lib.common_reg_cross_domain
  PORT MAP (
    in_rst      => mm_rst,
    in_clk      => mm_clk,
    in_dat      => mm_data_in,
    in_done     => OPEN,
    out_rst     => epcs_rst,
    out_clk     => epcs_clk,
    out_dat     => data_in,
    out_new     => OPEN
  );

  u_cross_data_out : ENTITY common_lib.common_reg_cross_domain
  PORT MAP (
    in_rst      => epcs_rst,
    in_clk      => epcs_clk,
    in_dat      => data_out,
    in_done     => OPEN,
    out_rst     => mm_rst,
    out_clk     => mm_clk,
    out_dat     => mm_data_out,
    out_new     => OPEN
  );
 
END rtl;

