###############################################################################
#
# Copyright (C) 2012
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

""" 

   One-Click prototype model "X"

   Remark: This example shows how a simple sysmtem of a block generator,
           a multiplexer and a databuffer can be modelled using threads
           with pipes. 
   
"""


###############################################################################
# System imports
import multiprocessing
import threading
import Queue
import time 


###############################################################################
# Functions
class base_component(object):
    def __init__(self, name, nof_inputs=0, nof_outputs=0, packetsize=10 ):
        self.name        = name
        self.nof_inputs  = nof_inputs
        self.nof_outputs = nof_outputs
        self.packetsize  = packetsize
        self.in_q        = []
        for i in range(nof_inputs):
            self.in_q.append(multiprocessing.Pipe(duplex=False))
        self.out_q = []
        print 'Instantiating', self.name
    
    #'>' operator: Connect
    def __gt__(self, other):
        return self.__connect__(other)
        
    # The __connect__ method fills the out_q attribute. 
    # out_q is a 3-dimensional list with pointers to the write-side of the
    # communication pipes: out_q[x][y][z], where
    #  x iterates over the fanout (the number of connected snk's).
    #  y iterates over the number of outputs. 
    #  z indexes the write-side of the pipe and is always [1]
    #
    def __connect__(self, other):
        if hasattr(other, 'in_q'):  
            # Create a list for all outputs. 
            other_in_q = []
            for j in xrange(self.nof_outputs):
              other_in_q.append(other.in_q[j])         
            # Add to the list to support multiple fanout
            self.out_q.append(other_in_q)
            print 'connecting', self.name, 'to', other.name 
            return other
        else:
            print 'Error: downstream component', other.name, 'does not have snk'

# Components inheriting the base component class
class blockGen(threading.Thread, base_component):
  
    def __init__(self, nof_outputs=1, data=[]):
      threading.Thread.__init__(self)
      base_component.__init__(self, 'blockGen', 0, nof_outputs, len(data[0][0]) )
      self.data = data
        
    def run(self):
      for i in xrange(len(self.data)):       # Iterate over the packets in time
        for h in xrange(len(self.out_q)):    # Iterate over the fanouts
          for j in xrange(self.nof_outputs): # Iterate over the outputs
            self.out_q[h][j][1].send(self.data[i][j])
            print "Sending packet " + str(i) + " on stream " + str(j) 
#            time.sleep(1)

class dp_mux(threading.Thread, base_component):
  
    def __init__(self, nof_inputs=2, nof_outputs=1, packetsize=10 ):
      threading.Thread.__init__(self)
      base_component.__init__(self, 'dp_mux', nof_inputs, nof_outputs, packetsize)
        
    def run(self):
      while True:
        data = []
        for i in xrange(self.nof_inputs):
          data.append(self.in_q[i][0].recv())
        for i in xrange(self.nof_inputs):
          for j in xrange(len(self.out_q)):     # Iterate over the fanouts
            self.out_q[j][0][1].send(data[i])   # DP_MUX has only one output.        


class dataBuffer(threading.Thread, base_component):
  
    def __init__(self, nof_inputs=1, nof_outputs=0, packetsize=10, nof_packets=12, index=0 ):
      threading.Thread.__init__(self)
      base_component.__init__(self, 'dataBuffer' + str(index), nof_inputs, nof_outputs, packetsize)
      self.nof_packets = nof_packets 
      self.index = index
        
    def run(self):
      data = []
      for j in xrange(self.nof_packets):
        for i in xrange(self.nof_inputs):
          data.append(self.in_q[i][0].recv())
      time.sleep(self.index)                   # Added a sleep to avoid clashing of the print output. 
      print self.name
      for i in data:
          print i

# Some definitions
packetsize = 10
bg_nof_outputs = 2
nof_transm_packets = 12
nof_receiv_packets = 24

# Create data for the block generator
#
# bg_data=[h][i][j]
#   h = [0:nof_transm_packets-1]
#   i = [0:nof_outputs-1]
#   j = [0:packetsize-1]
#
# The size of bg_data determines the various paramters of the model. 

bg_data = []
for h in xrange(nof_transm_packets):
    packets = []
    for i in xrange(bg_nof_outputs):
        packet = []
        for j in xrange(packetsize):
            packet.append(j+i*packetsize+h*bg_nof_outputs*packetsize)
        packets.append(packet)
    bg_data.append(packets)

# Instantiate the components for the system
bg  = blockGen(bg_nof_outputs, bg_data)
mux = dp_mux(bg_nof_outputs, 1, packetsize)
db1 = dataBuffer(1, 0 , packetsize, nof_receiv_packets, 1)         
db2 = dataBuffer(1, 0 , packetsize, nof_receiv_packets, 2)  

# Connect the components
bg > mux > db1
mux > db2

# Start the threads 
bg.start()
mux.start()
db1.start()
db2.start()
        
